"""
MailBackend
~~~~~~~~~~~~~~
reads smtp config from ini file and sends out emails as directed by main

    :copyright: 2015 by Matthias Kauer
    :license: BSD
"""

from smtplib import SMTP_SSL
import os
import ConfigParser

class MailBackend(object):
    def __init__(self, scriptpath):
        self.smtp = SMTP_SSL()
        self.cfg = ConfigParser.SafeConfigParser()
        fname = os.path.realpath(scriptpath)

        try:
            self.cfg.read(fname)
            self.server = self.cfg.get('Mail', 'smtpserver')
            self.port = self.cfg.get('Mail', 'smtpport')
            self.mailuser = self.cfg.get('Mail','user')
            self.mailpass = self.cfg.get('Mail','pass')
            self.fromaddr = self.cfg.get('Mail', 'fromaddr')
            self.loggedin = False
        except Exception, e:
            import sys
            sys.exit( "%s appears not to be an .ini file with the appropriate sections.\n \
Call 'croncoat --ini' for an example layout of the .ini file" %scriptpath )

    def sendmail(self, emailMsg):
        if(not self.loggedin):
            emailMsg['From']=self.fromaddr
            
            self.smtp.connect(self.server, self.port)
            self.smtp.login(self.mailuser, self.mailpass)
            self.loggedin = True

        self.smtp.sendmail(emailMsg['From'], emailMsg['To'], emailMsg.as_string())

    def __exit__(self):
        print("exiting MailBackend")
        self.smtp.quit()






"""
Helper
~~~~~~~~~~~~~~
contains helper functionality: trimming messages, parsing time

    :copyright (cronwrap): 2010 by Plurk
    :license: BSD
"""

import re
from datetime import datetime, timedelta

class Helper(object):
    '''
    classdocs
    '''


    def __init__(self):
        '''
        Constructor
        '''
#         self.sname = scriptname
        pass
    
    @staticmethod
    def parse_time_to_secs(arg_time):
        #Parse sys_args.time
        max_time = arg_time
        sp = re.match("(\d+)([hms])", max_time).groups()
        t_val, t_unit = int(sp[0]), sp[1]

        #Convert time to seconds
        if t_unit == 'h':
            t_val = t_val * 60 * 60
        elif t_unit == 'm':
            t_val = t_val * 60

        return t_val
    
    @staticmethod
    def trim_if_needed(txt, max_length=10000):
        if len(txt) > max_length:
            return '... START TRUNCATED...\n%s' % txt[-max_length:]
        else:
            return txt
        
    @staticmethod
    def render_email_template(title, sys_args, cmd):
        result_str = []

        result_str.append(title)
        result_str.append('%s\n' % sys_args.cmd)

        result_str.append('COMMAND STARTED:')
        result_str.append('%s UTC\n' % (datetime.utcnow() -
                                        timedelta(seconds=int(cmd.run_time))))

        result_str.append('COMMAND FINISHED:')
        result_str.append('%s UTC\n' % datetime.utcnow())

        result_str.append('COMMAND RAN FOR:')
        hours = cmd.run_time / 60 / 60
        if cmd.run_time < 60:
            hours = 0
        result_str.append('%d seconds (%.2f hours)\n' % (cmd.run_time, hours))

        result_str.append("COMMAND'S TIMEOUT IS SET AT:")
        result_str.append('%s\n' % sys_args.time)

        result_str.append('RETURN CODE WAS:')
        result_str.append('%s\n' % cmd.returncode)

        result_str.append('ERROR OUTPUT:')
        result_str.append('%s\n' % Helper.trim_if_needed(cmd.stderr))

        result_str.append('STANDARD OUTPUT:')
        result_str.append('%s' % Helper.trim_if_needed(cmd.stdout))

        return '\n'.join(result_str)
    
    @staticmethod
    def is_time_exceeded(sys_args, cmd):
        """Returns `True` if the command's run time has exceeded the maximum
        run time specified in command arguments. Else `False  is returned."""
        cmd_time = int(cmd.run_time)

        t_val = Helper.parse_time_to_secs(sys_args.time)
        return cmd_time > t_val



#!/usr/bin/env python

"""
croncoat
~~~~~~~~~~~~~~
A cron job wrapper that wraps jobs and enables better error reporting and command timeouts.
croncoat added: python mail handling, better subject lines, timeout kills ongoing tasks,
object-oriented design

    :copyright (cronwrap): 2010 by Plurk
    :copyright: 2015 by Matthias Kauer
    :license: BSD
"""
#  __VERSION__ = '0.2'

#  import re
import argparse
import sys
import os
from croncoat.cc.cronwrapper import CronWrapper
from croncoat import __version__, __scriptname__

class MyParser(argparse.ArgumentParser):
    """custom parser class so I can output help whenever wrong / bad / no arguments are supplied"""
#  http://stackoverflow.com/questions/4042452/display-help-message-with-python-argparse-when-script-is-called-without-any-argu
    def error(self, message):
        sys.stderr.write('error: %s\n\n' % message)
        self.print_help()
        sys.exit(2)


desc_str =  (
"""Wrap cron jobs for better error email error reporting with command timeouts.
Version {}
You must create a config file (~/.{}.ini by default) to store smtp server data.
Ideally this would be readable only by you.
To output a config skeleton, use {} --print-ini

Usage examples:
===============
Send test email:
{} -e test@domain.org

Send email after killing a script that takes longer than 5s
{} -t 5s -c 'sleep 10s' -e test@domain.org

Print to stdout after catching error in script;
Note: this won't work with exit(1) b/c no real shell here
{} -c 'python -c import sys; sys.exit(1)'

Print no output for successful command
{} -c 'ls -la'
Print output of successful command
{} -v -c 'ls -la'""").format(__version__, *([__scriptname__] * 7))

#  if __name__ == '__main__':
def main(input_args=None):
    parser = MyParser(prog=__scriptname__, description=desc_str, formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('-c', '--cmd',
                        help='Run a command. Could be `%s -c "ls -la"`. No command => test email is sent.' % __scriptname__
                        )

    parser.add_argument('-e', '--emails',
                        help='Send email to the following addresses if the command crashes or exceeds timeout. '
                        "Uses Python's email library to send emails (therefore no user names"
                        "unlike original cronwrap). "
                        "If this is not set, only output to stdout."
                        )

    parser.add_argument('-t', '--time',
                        help='Set the maximum running time. '
                        'If this time is reached, the script will be killed and an alert email will be sent. '
                        "If the script is killed stdout/stderr cannot be captured at this time! "
                        "The default is 1 hour `-t 1h`. Possible values include: `-t 2h`,`-t 5m`, `-t 30s`."
                        )

    parser.add_argument('--print-ini', '--print-config', nargs='?', default=False,
                        help='Print the configuration file format. '
                            'This can be redirected to a file name to have a config skeleton.'
                        )

    parser.add_argument('--config', '--ini', '-i', default=False,
                        help='use an .ini file with custom name and path  '
                        "(not the default .croncoat.ini in users' home directory"
                        )

    parser.add_argument('-v', '--verbose',
                        nargs='?', default=False,
                        help='Will send an email / print to stdout even on successful run.')

    sys_args = parser.parse_args(input_args)
    if(sys_args.print_ini is not False):
        CronWrapper.print_ini(__scriptname__)
    elif(sys_args.cmd is None and sys_args.emails is None):
        sys.stderr.write('error: neither command nor email was supplied;'
             'email alone => test email, command alone => console output\n\n')
        parser.print_help()
        sys.exit(1)
    else:
        if sys_args.config is not False:
            configpath = os.path.realpath(sys_args.config)
        else: # default
            configpath = os.path.expanduser("~/.%s.ini" % __scriptname__)
        if os.path.exists(configpath) and os.path.isfile(configpath):
            #  print(configpath)
            cwrap = CronWrapper(sys_args, __scriptname__, configpath)
            cwrap.run()
        else:
            sys.exit("no config file detected at %s" %configpath)





"""
croncoat tests
~~~~~~~~~~~~~~

    :copyright: 2015 by Matthias Kauer
    :license: BSD
"""
import unittest
from croncoat.scripts.ccscript import main
from croncoat.cc.cronwrapper import CronWrapper
from croncoat import __scriptname__
import os

class Test(unittest.TestCase):
    
    def setUp(self):
        f = open('temp.ini', 'w')
        f.write(CronWrapper._ini_string(__scriptname__))
        f.close
        self.base_arg = ['-i', 'temp.ini']
        
    def tearDown(self):
        if os.path.exists('temp.ini'):
            os.remove('temp.ini')
        
    def test_expiring(self):
        input_args = self.base_arg + ["-t", "1s", "-c", "sleep 2"]
        with self.assertRaises(SystemExit):
            main(input_args)

        input_args = self.base_arg + ["-t", "3s", "-c", "sleep 2"]
        main(input_args) #does not raise

    def test_errorcode(self):
        input_args = self.base_arg + ['-c', 'python -c "import sys; sys.exit(1)"']
        with self.assertRaises(SystemExit):
            main(input_args)

    def test_success(self):
        input_args = self.base_arg + ['-c', 'ls -la']
        main(input_args)

        input_args = self.base_arg +  ['-v', '-c', 'ls -la']
        main(input_args)

if __name__ == "__main__":
    unittest.main()




"""
ExpiringCommand
~~~~~~~~~~~~~~
Signal based implementation that starts a command and kills it once a timeout
is reached.
Caveat: Currently does not save stdout when command is killed

    :copyright: 2015 by Matthias Kauer
    :license: BSD
"""

import time
import signal
import subprocess
import shlex
from croncoat.cc.helper import Helper

class Alarm(Exception):
    pass

def alarm_handler(signum, frame):
    raise Alarm

class ExpiringCommand(object):
    #http://stackoverflow.com/questions/1191374/subprocess-with-timeout
    
    def savereturn(self):
        self.run_time = time.time() - self.start_time
        self.returncode = self.p.returncode;

        #  self.stderr = "asdf"
        #  self.stdout = self.p.stdout.read()
        #  with open(self.outfile) as fout:
        #  with open(self.outfile) as fout, open(self.errfile) as ferr:
        #      self.stdout = fout.read()
        #      self.stderr = ferr.read()
        #      print("ExpSignal; stdout: %s %s" % (self.stdout, self.stderr))

    def __init__(self, command, timeout):
#         self.cmd = command.split()
        #  print(command)
        self.cmd = shlex.split(command)
        self.timeout = Helper.parse_time_to_secs(timeout)
        self.start_time = time.time()

        self.outfile = 'stdout.txt'
        self.errfile = 'stderr.txt'

        #  self.p = subprocess.Popen(self.cmd,stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=True)

    def Run(self):
        #  print("cmd: ", self.cmd)
        self.p = subprocess.Popen(self.cmd,stderr=subprocess.PIPE, stdout=subprocess.PIPE, shell=False,
                universal_newlines=True)
        #  with open(self.outfile, 'w') as fout, open(self.errfile, 'w') as ferr:
            #  self.p = subprocess.Popen(self.cmd,stderr=ferr, stdout=fout, shell=False)

        signal.signal(signal.SIGALRM, alarm_handler)
        signal.alarm(self.timeout)

        self.stderr = "Can't capture output if terminated early (sorry, but I just spent 4h trying to get this)\n"
        self.stderr += "Best option at this points seems to be the queue implementation here: \n"
        self.stderr += "http://stackoverflow.com/questions/375427/non-blocking-read-on-a-subprocess-pipe-in-python\n"
        self.stdout = self.stderr

        try:
            self.p.wait()
            self.returncode = self.p.returncode
            self.stderr = self.p.stderr.read()
            self.stdout = self.p.stdout.read()
        except Alarm:
            #  print("cmd %s timed out; killing now" % (self.cmd))
            self.p.terminate()

            #this didn't work
            #  self.stdout = []
            #  self.stderr = []
            #  for line in self.p.stdout:
            #      self.stdout.append(line)
            #  for line in self.p.stderr:
            #      self.stderr.append(line)


            #  self.returncode = -1
        finally:
            #  if not self.stdout:
            #      self.stdout = self.p.stdout.read()
            #  if not self.stderr:
            #      self.stderr = self.p.stderr.read()
            self.savereturn()


